#!/usr/bin/env bash

script_name="$(basename $(realpath $0))"
working_dir="$(dirname $(realpath $0))"


if [[ -f "${working_dir}/conf/borg2-backup-nc.conf" ]] ; then
   conf_file="${working_dir}/conf/borg2-backup-nc.conf"
else
   conf_file="${working_dir}/conf/borg2-backup.conf"
fi

LOCK_DIR="/tmp/${script_name%%.*}.LOCK"
log_file="${LOCK_DIR}/${script_name%%.*}.log"



# ----------
# Base Function(s)
# ----------

clean_up() {

   # Perform program exit housekeeping
   rm -rf "$LOCK_DIR"
   blank_line
   exit $1
}

echononl(){
   if $terminal ; then

      echo X\\c > /tmp/shprompt$$
      if [ `wc -c /tmp/shprompt$$ | awk '{print $1}'` -eq 1 ]; then
         echo -e -n "   $*\\c" 1>&2
      else
         echo -e -n "   $*" 1>&2
      fi
      rm /tmp/shprompt$$

   fi

}

fatal(){
   echo ""
   if $terminal ; then
      echo -e "   [ \033[31m\033[1mFatal\033[m ]: $*"
      echo ""
      echo -e "   \033[31m\033[1mScript was interupted\033[m!"
   else
      echo "   [ Fatal ]: $*"
      echo ""
      echo "              Script was terminated...."
   fi
   echo ""
   clean_up 1
}
error (){
   echo ""
   if $terminal ; then
      echo -e "   [ \033[31m\033[1mError\033[m ]: $*"
   else
      echo "[ Error ]: $*"
   fi
}

warn (){
   echo ""
   if $terminal ; then
      echo -e "   [ \033[33m\033[1mWarning\033[m ]: $*"
   else
      echo "[ Warning ]: $*"
   fi
   echo ""
}

warn_only_terminal () {
   if $terminal ; then
      echo ""
      echo -e "\t[ \033[33m\033[1mWarning\033[m ]: $*"
      echo ""
   fi
}

info (){
   if $terminal ; then
      echo ""
      echo -e "   [ \033[32m\033[1mInfo\033[m ]  $*"
      echo ""
   fi
}

ok (){
   if $terminal ; then
      echo ""
      echo -e "   [ \033[32m\033[1mOk\033[m ]    $*"
      echo ""
   fi
}

echo_done() {
   if $terminal ; then
      echo -e "\033[75G[ \033[32mdone\033[m ]"
   fi
}
echo_ok() {
   if $terminal ; then
      echo -e "\033[75G[ \033[32mok\033[m ]"
   fi
}
echo_warn() {
   if $terminal ; then
      echo -e "\033[75G[ \033[33mwarn\033[m ]"
   fi
}
echo_failed(){
   if $terminal ; then
      echo -e "\033[75G[ \033[1;31mfailed\033[m ]"
   fi
}
echo_skipped() {
   if $terminal ; then
      echo -e "\033[75G[ \033[90m\033[1mskipped\033[m ]"
   fi
}
echo_wait(){
   if $terminal ; then
      echo -en "\033[75G[ \033[5m\033[1m...\033[m ]"
   fi
}

trim() {
    local var="$*"
    var="${var#"${var%%[![:space:]]*}"}"   # remove leading whitespace characters
    var="${var%"${var##*[![:space:]]}"}"   # remove trailing whitespace characters
    echo -n "$var"
}

blank_line() {
   if $terminal ; then
      echo ""
   fi
}

# ----------
# - Some checks ..
# ----------

# - Running in a terminal?
# -
if [[ -t 1 ]] ; then
   terminal=true
else
   terminal=false
fi


# -------------
# - Job is already running?
# -------------

# - If job already runs, stop execution..
# -
if mkdir "$LOCK_DIR" 2> /dev/null ; then

   ## - Remove lockdir when the script finishes, or when it receives a signal
   trap "clean_up 1" SIGHUP SIGINT SIGTERM

else

   datum="$(date +"%d.%m.%Y %H:%M")"

   msg="[ Error ]: A previos instance of \"`basename $0`\" seems already be running.\n\n           Exiting now.."

   echo ""
   echo "[ Error ]: A previos instance of that script \"`basename $0`\" seems already be running."
   echo ""
   echo -e "           Exiting now.."
   echo ""

   for _email in ${alert_email_arr[@]} ; do
      echo -e "To:${_email}\n${content_type}\nSubject:Error cronjob `basename $0` -- $datum\n${msg}\n" \
         | sendmail -F "Error `hostname -f`" -f $sender_address $_email
   done

   exit 1

fi


# ==========
# - Begin Main Script
# ==========

# ----------
# - Headline
# ----------

if $terminal ; then
   echo ""
   echo -e "\033[1m----------\033[m"
   echo -e "\033[32m\033[1mRunning script \033[m\033[1m$script_name\033[32m .. \033[m"
   echo -e "\033[1m----------\033[m"
   echo ""
fi

# ----------
# Default values
# ----------

DEFAULT_SSH_USER="borg"
DEFAULT_SSH_PORT=22
DEFAULT_BORG_RSH='ssh -i /root/.ssh/id_ed25519-borg-backup'

DEFAULT_BORG_FUSE_MOUNT_DIR="/mnt/borg-fuse-mount"

# ----------
#  Read Configurations from $conf_file
# ----------


if [[ -f "$conf_file" ]]; then
   source "$conf_file"
else
   fatal  "No configuration file '$conf_file' present!"
fi


# ----------
# Set Parameter values
# ----------

#if [[ -z "${BORG_PASSPHRASE}" ]] ; then
#   fatal "Borg Passphrase (BORG_PASSPHRASE) NOT found!"
#fi

if [[ -z "${BORG_HOST}" ]] ; then
   fatal "Backup Hosth (var BORG_HOST) not given!"
fi

if [[ -z "${BORG_REPO}" ]] ; then
   fatal "Borg Repository Path (var BORG_REPO) not given!"
fi

[[ -z "${HTTP_USER}" ]] && HTTP_USER="${DEFAULT_HTTP_USER}"

[[ -z "${SSH_USER}" ]] && SSH_USER="${DEFAULT_SSH_USER}"

[[ -z "${SSH_PORT}" ]] && SSH_PORT="${DEFAULT_SSH_PORT}"

[[ -z "${BORG_RSH}" ]] && BORG_RSH="${DEFAULT_BORG_RSH}"

[[ -z "${BORG_FUSE_MOUNT_DIR}" ]] && BORG_FUSE_MOUNT_DIR="${DEFAULT_BORG_FUSE_MOUNT_DIR}"

BORG_URL="ssh://${SSH_USER}@${BORG_HOST}:${SSH_PORT}${BORG_REPO}"

export BORG_PASSPHRASE
export BORG_REPO
export BORG_RSH

# ---
# Some automatic “answerers” (if set, they automatically answer confirmation questions):
# ---

# For 'Warning: The repository at location … was previously located at ..'
#
BORG_RELOCATED_REPO_ACCESS_IS_OK=yes
export BORG_RELOCATED_REPO_ACCESS_IS_OK
blank_line


LIST_OF_FILES_OR_DIRECTORIES=
echo -e "\033[32m--\033[m"
echo ""
echo "Specify a space-separated list of files or directories to be part of the mounted filesystem."
echo ""
echo -e "   Type \"\033[33mall\033[m\" or leave emty to mount the full archive"
echo ""
echononl "List of files or directories: "
read LIST_OF_FILES_OR_DIRECTORIES
if [[ "X${LIST_OF_FILES_OR_DIRECTORIES}" = "X" ]] || [[ "$(trim ${LIST_OF_FILES_OR_DIRECTORIES,,})" = 'all' ]] ; then
   LIST_OF_FILES_OR_DIRECTORIES=""
fi

blank_line

echo -e "\033[32m--\033[m"
echo ""

# List Borg Backups
#
(
   source ${working_dir}/borg2-list-backups.sh
)

echo ""
echo -e "   \033[33m--------------------------------------------\033[m"
echo "   Which Borg Archiv shoul be mounted?"
echo ""
echo -e "      \033[33mNote:\033[m it is an read-only fuse filesystem!"
echo ""
echo "   Give a Borg Backup ID from above list"
echo ""
echo -n "   Borg Backup ID: "
read BORG_BACKUP_ID
echo -e "   \033[33m--------------------------------------------\033[m"
echo ""


blank_line


echo ""
echo ""
echo -e "\t\033[32mStart script for (readonly) mounting repository archiv '${BORG_BACKUP_ID}'..\033[m"
echo ""
echo -e "\tBORG_HOST......................: ${BORG_HOST}"
echo -e "\tBORG_REPO......................: ${BORG_REPO}"
echo ""
echo -e "\tBORG_PASSPHRASE................: ${BORG_PASSPHRASE}"
echo ""
echo -e "\tSSH_USER ......................: ${SSH_USER}"
echo -e "\tSSH_PORT.......................: ${SSH_PORT}"
echo -e "\tBORG_RSH.......................: ${BORG_RSH}"
echo ""
echo -e "\tBORG_URL.......................: ${BORG_URL}"
echo ""
echo -e "\tBORG_FUSE_MOUNT_DIR............: ${BORG_FUSE_MOUNT_DIR}"
if [[ -z "${LIST_OF_FILES_OR_DIRECTORIES}" ]]; then
   echo -e "\tLIST_OF_FILES_OR_DIRECTORIES...: \033[33mMount the full archive\033[m"
else
   echo -e "\tLIST_OF_FILES_OR_DIRECTORIES...: ${LIST_OF_FILES_OR_DIRECTORIES}"
fi
echo ""
echo ""

declare -i _CHOOSE
echo -e "   Ho to continue?"
echo ""
echo -e "   \033[1m[1] Mount files from repository $(basename "${BORG_REPO}") archive ${BORG_BACKUP_ID}\033[m"
echo -e "   [2] Only print out the commands for mounting files/folders"
echo -e "   [3] Exit borg2 mount script."
echo ""
echo "   Type a number or press <RETURN> to choose highlighted value"
echo ""
echononl "Eingabe: "
while [[ -z "${_CHOOSE}" ]]; do
   read OPTION
   [[ -z "$OPTION" ]] && OPTION=1
   case $OPTION in
      1) _CHOOSE=1
         ;;
      2) _CHOOSE=2
         ;;
      3) _CHOOSE=3
         ;;
      *)
         echo ""
         echo -e "\tFalsche Eingabe ! [1/2/3] or type <RETURN> for the highlited default"
         echo ""
         echononl "Eingabe: "
         ;;
   esac
done

if [[ ${_CHOOSE} -gt 2 ]] ; then
   echo ""
   clean_up 0
elif [[ ${_CHOOSE} -gt 1 ]] ; then

   _MOUNT_DIR_BACKUP_HOST="${BORG_FUSE_MOUNT_DIR}/$(basename "${BORG_REPO}")"

   echo -e "\033[40m\033[37m
-----------------------------------------------------------------------------------------

export BORG_PASSPHRASE=\"${BORG_PASSPHRASE}\"
export BORG_REPO=\"${BORG_REPO}\"
export BORG_RSH=\"${BORG_RSH}\"
export BORG_RELOCATED_REPO_ACCESS_IS_OK=yes

mkdir -p \"${BORG_FUSE_MOUNT_DIR}\"

borg2 --remote-path borg2 --repo ${BORG_URL} mount -a ${BORG_BACKUP_ID} ${BORG_FUSE_MOUNT_DIR} ${LIST_OF_FILES_OR_DIRECTORIES}


On Backup Host type:
====================

export BORG_PASSPHRASE=\"${BORG_PASSPHRASE}\"
export BORG_REPO=\"${BORG_REPO}\"
export BORG_RELOCATED_REPO_ACCESS_IS_OK=yes

mkdir -p \"${_MOUNT_DIR_BACKUP_HOST}\"

borg2 --remote-path borg2 --repo ${BORG_REPO} mount -a ${BORG_BACKUP_ID} ${_MOUNT_DIR_BACKUP_HOST} ${LIST_OF_FILES_OR_DIRECTORIES}


You can find the mounted fuse archive at;

   ${_MOUNT_DIR_BACKUP_HOST}/${BORG_BACKUP_ID}


In Order to unmount repository archive type:

   umount ${_MOUNT_DIR_BACKUP_HOST}

-----------------------------------------------------------------------------------------
\033[m\n"

clean_up 0
fi

echononl "Create Restore Directory '${BORG_FUSE_MOUNT_DIR}'."
if [ ! -d "${BORG_FUSE_MOUNT_DIR}" ]; then
   mkdir -p "${BORG_FUSE_MOUNT_DIR}" > ${log_file} 2>&1
   if [[ $? -gt 0 ]] ; then
      echo_failed
      error "$(cat "${log_file}")"

      fatal "No restore directory '${BORG_FUSE_MOUNT_DIR}' present!"

   else
      echo_ok
   fi
else
   echo_skipped
fi

blank_line

_MOUNT_DIR_BACKUP_HOST="${BORG_FUSE_MOUNT_DIR}/$(basename "${BORG_REPO}")"

echo -e "\033[40m\033[37m
-----------------------------------------------------------------------------------------

export BORG_PASSPHRASE=\"${BORG_PASSPHRASE}\"
export BORG_REPO=\"${BORG_REPO}\"
export BORG_RSH=\"${BORG_RSH}\"
export BORG_RELOCATED_REPO_ACCESS_IS_OK=yes

mkdir -p \"${BORG_FUSE_MOUNT_DIR}\"

borg2 --remote-path borg2 --repo ${BORG_URL} mount -a ${BORG_BACKUP_ID} ${BORG_FUSE_MOUNT_DIR} ${LIST_OF_FILES_OR_DIRECTORIES}


On Backup Host type:
====================

export BORG_PASSPHRASE=\"${BORG_PASSPHRASE}\"
export BORG_REPO=\"${BORG_REPO}\"
export BORG_RELOCATED_REPO_ACCESS_IS_OK=yes

mkdir -p \"${_MOUNT_DIR_BACKUP_HOST}\"

borg2 --remote-path borg2 --repo ${BORG_REPO} mount -a ${BORG_BACKUP_ID} ${_MOUNT_DIR_BACKUP_HOST} ${LIST_OF_FILES_OR_DIRECTORIES}


You can find the mounted fuse archive at;

   ${_MOUNT_DIR_BACKUP_HOST}/${BORG_BACKUP_ID}


In Order to unmount repository archive type:

   umount ${_MOUNT_DIR_BACKUP_HOST}

-----------------------------------------------------------------------------------------
\033[m\n"

blank_line

echo "Mount archive '$(basename "${BORG_REPO}")::${BORG_BACKUP_ID}' to"
echononl "      '${BORG_FUSE_MOUNT_DIR}'.."

borg2 --remote-path borg2 --repo ${BORG_URL} mount -a ${BORG_BACKUP_ID} ${BORG_FUSE_MOUNT_DIR} ${LIST_OF_FILES_OR_DIRECTORIES} > "${log_file}" 2>&1

if [[ $? -gt 0 ]] ; then
   echo_failed
   error "$(cat "${log_file}")"
else
   echo_ok
fi


blank_line

info "You can find the readonly mounted files from repository archive '${BORG_BACKUP_ID}' at directory:

             \033[1m${BORG_FUSE_MOUNT_DIR}/${BORG_BACKUP_ID}\033[m"

clean_up 0
